import pytest

from .hash_cleaned_str import clean_str, hash_cleaned_str


def test_hash_cleaned_str_should_ignore_punctuation():
    strings_should_hash_to_identical = ["this", "This"]

    punctuation = r"!().,:;/"
    strings_should_hash_to_identical += [
        f"{s}{punctuation}" for s in strings_should_hash_to_identical
    ]

    assert (
        len({hash_cleaned_str(s) for s in strings_should_hash_to_identical})
        == 1
    )


def test_hash_cleaned_str_should_remove_html_tags():
    strings_should_hash_to_identical = ["<p>Test</p>", "Test"]

    assert hash_cleaned_str(
        strings_should_hash_to_identical[0]
    ) == hash_cleaned_str(strings_should_hash_to_identical[1])


@pytest.mark.parametrize(
    ("input_str", "expected"), [("Is <2, but >4", "is2but4")]
)
def test_str_cleaner(input_str: str, expected: str):
    assert clean_str(input_str) == clean_str(expected)



from collections.abc import Sequence
from typing import Protocol

from .document import Document
from .document_source import BaseDocumentSource
from .extractors.extractor import BasePromptExtractor
from .prompts.prompt import BasePrompt


class BasePromptSource(Protocol):
    def get_prompts(self) -> Sequence[BasePrompt]:
        ...


class DocumentPromptSource(BasePromptSource):
    def __init__(
        self,
        document_ingester: BaseDocumentSource,
        prompt_extractors: Sequence[BasePromptExtractor],
    ):
        self._document_ingester = document_ingester
        self._prompt_extractors = prompt_extractors

    def _get_prompts_from_document(
        self, document: Document
    ) -> Sequence[BasePrompt]:
        prompts: list[BasePrompt] = []
        for extractor in self._prompt_extractors:
            extractor_prompts = list(extractor.extract_prompts(document))
            prompts += extractor_prompts

        return prompts

    def get_prompts(self) -> Sequence[BasePrompt]:
        documents = self._document_ingester.get_documents()

        prompts = [self._get_prompts_from_document(doc) for doc in documents]

        return [item for sublist in prompts for item in sublist]



from collections.abc import Sequence
from dataclasses import dataclass, field

from .anki_prompt_cloze import AnkiCloze
from .anki_prompt_qa import AnkiQA


def fake_tag_factory() -> Sequence[str]:
    return ["FakeTag"]


@dataclass(frozen=True)
class FakeAnkiCloze(AnkiCloze):
    text: str = "FakeText"
    base_deck: str = "FakeBaseDeck"
    tags: Sequence[str] = field(default_factory=fake_tag_factory)
    css: str = "FakeCSS"
    uuid: int = 0


@dataclass(frozen=True)
class FakeAnkiQA(AnkiQA):
    base_deck: str = "FakeBaseDeck"
    tags: Sequence[str] = field(default_factory=fake_tag_factory)
    question: str = "FakeQuestion"
    answer: str = "FakeAnswer"
    css: str = "FakeCSS"
    uuid: int = 0


def test_ankiqa_deck_inference():
    card = FakeAnkiQA(tags=["anki/deck/Subdeck"])
    assert card.deck == "FakeBaseDeck::Subdeck"


def test_formatting():
    card = FakeAnkiQA(question="Q. This is a _question_?")
    note = card.to_genanki_note()
    assert note.fields[0] == "<p>Q. This is a <em>question</em>?</p>"  # type: ignore



from collections.abc import Sequence
from dataclasses import dataclass

from ...utils.hash_cleaned_str import hash_cleaned_str, int_hash_str
from ..document import Document
from .prompt import BasePrompt


@dataclass(frozen=True)
class QAPrompt(BasePrompt):
    question: str
    answer: str

    @property
    def scheduling_uid(self) -> int:
        return hash_cleaned_str(f"{self.question}_{self.answer}")

    @property
    def update_uid(self) -> int:
        return int_hash_str(f"{self.question}_{self.answer}_{self.tags}")

    @property
    def tags(self) -> Sequence[str]:
        return ()


@dataclass(frozen=True)
class QAWithoutDoc(QAPrompt):
    add_tags: Sequence[str]

    @property
    def tags(self) -> Sequence[str]:
        return self.add_tags


@dataclass(frozen=True)
class QAFromDoc(QAPrompt):
    parent_doc: Document
    line_nr: int

    @property
    def tags(self) -> Sequence[str]:
        return self.parent_doc.tags



from collections.abc import Sequence
from dataclasses import dataclass

from ...utils.hash_cleaned_str import hash_cleaned_str, int_hash_str
from ..document import Document
from .prompt import BasePrompt


@dataclass(frozen=True)
class ClozePrompt(BasePrompt):
    text: str

    @property
    def scheduling_uid(self) -> int:
        return hash_cleaned_str(self.text)

    @property
    def update_uid(self) -> int:
        return int_hash_str(f"{(self.text)}{self.tags}")

    @property
    def tags(self) -> Sequence[str]:
        return ()


@dataclass(frozen=True)
class ClozeWithoutDoc(ClozePrompt):
    add_tags: Sequence[str]

    @property
    def tags(self) -> Sequence[str]:
        return self.add_tags


@dataclass(frozen=True)
class ClozeFromDoc(ClozePrompt):
    text: str
    source_doc: Document

    @property
    def scheduling_uid(self) -> int:
        return int_hash_str(self.text)

    @property
    def tags(self) -> Sequence[str]:
        return self.source_doc.tags



import hashlib
from collections.abc import Callable

from bs4 import BeautifulSoup


def remove_spaces(text: str) -> str:
    """Remove spaces from a string."""
    return text.replace(" ", "")


def remove_html_tags(text: str) -> str:
    clean_text = BeautifulSoup(text, "html.parser").text
    return clean_text


def remove_punctuation(text: str) -> str:
    """Clean string before hashing, so changes to spacing, punctuation, newlines etc. do not affect the hash."""
    lowered = text.lower()

    punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`|~"""
    cleaned = lowered.translate(str.maketrans("", "", punctuation))

    return cleaned


def clean_str(input_str: str) -> str:
    """Clean string before hashing, so changes to spacing, punctuation, newlines etc. do not affect the hash."""
    cleaned = input_str

    for cleaner in [remove_html_tags, remove_punctuation, remove_spaces]:
        cleaned = cleaner(cleaned)

    return cleaned


def int_hash_str(input_string: str, max_length: int = 10) -> int:
    # Convert the string to bytes
    bytes_string = input_string.encode()

    # Create a SHA256 hash object
    hash_object = hashlib.sha256(bytes_string)

    # Hex digest the hash and convert it to an integer
    unique_int = int(hash_object.hexdigest(), 16)
    shortened = unique_int % 10**max_length

    return shortened


def hash_cleaned_str(
    input_str: str, cleaner: Callable[[str], str] = clean_str
) -> int:
    """Hash a string after cleaning it."""
    hashed = int_hash_str(cleaner(input_str))
    return hashed



import logging
from functools import partial
from pathlib import Path
from typing import Annotated, Optional

import sentry_sdk
import typer

from .destination.ankiconnect.anki_converter import AnkiPromptConverter
from .destination.ankiconnect.ankiconnect_gateway import (
    ANKICONNECT_URL,
    AnkiConnectGateway,
)
from .destination.destination import PushPrompts
from .destination.destination_ankiconnect import AnkiConnectDestination
from .destination.destination_dryrun import DryRunDestination
from .diff_determiner import PromptDiffDeterminer
from .environment import get_env, host_input_dir, in_docker
from .source.document_source import MarkdownDocumentSource
from .source.extractors.extractor_cloze import ClozePromptExtractor
from .source.extractors.extractor_qa import QAPromptExtractor
from .source.source import DocumentPromptSource

log = logging.getLogger(__name__)

app = typer.Typer()


def main(
    base_deck: str,
    input_dir: Path,
    max_deletions_per_run: int,
    dry_run: bool,
    push_all: bool = False,
):
    source_prompts = DocumentPromptSource(
        document_ingester=MarkdownDocumentSource(directory=input_dir),
        prompt_extractors=[
            QAPromptExtractor(question_prefix="Q.", answer_prefix="A."),
            ClozePromptExtractor(),
        ],
    ).get_prompts()

    dest_class = AnkiConnectDestination if not dry_run else DryRunDestination
    destination = dest_class(
        gateway=AnkiConnectGateway(
            ankiconnect_url=ANKICONNECT_URL,
            base_deck=base_deck,
            tmp_read_dir=host_input_dir() if in_docker() else input_dir,
            tmp_write_dir=input_dir,
            max_deletions_per_run=max_deletions_per_run,
            max_wait_seconds=0,
        ),
        prompt_converter=AnkiPromptConverter(
            base_deck=base_deck,
            card_css=Path(
                "memium/destination/ankiconnect/default_styling.css"
            ).read_text(),
        ),
    )
    if not push_all:
        destination_prompts = destination.get_all_prompts()
        update_commands = PromptDiffDeterminer().sync(
            source_prompts=source_prompts,
            destination_prompts=destination_prompts,
        )
        destination.update(commands=update_commands)
    else:
        destination.update([PushPrompts(prompts=source_prompts)])


@app.command()
def cli(
    input_dir: Annotated[
        Path,
        typer.Option(
            help="Directory containing markdown prompts to sync to Anki.",
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
            writable=True,
        ),
    ],
    watch: Annotated[
        Optional[int],  # noqa: UP007
        typer.Option(help="Keep running, updating Anki deck every 15 seconds"),
    ] = None,
    deck_name: Annotated[
        str,
        typer.Option(help="Anki path to deck, e.g. 'Parent deck::Child deck'"),
    ] = "Memium",
    max_deletions_per_run: Annotated[
        int,
        typer.Option(
            help="Maximum number of cards to delete per sync to avoid unintentional deletions. If exceeded, raises error."
        ),
    ] = 50,
    push_all: Annotated[
        bool,
        typer.Option(
            help="Push all prompts to Anki, whether or not they exist in Anki. Useful if you have e.g. changed your CSS template. Note that this does not change scheduling, nor delete prompts that no longer exist in your markdown."
        ),
    ] = False,
    dry_run: Annotated[
        bool,
        typer.Option(
            help="Don't update via AnkiConnect, just log what would happen"
        ),
    ] = False,
    sentry_dsn: Annotated[
        Optional[str],  # noqa: UP007
        typer.Option(
            help="Sentry DSN for error reporting. If not provided, no error reporting will be done."
        ),
    ] = None,
    skip_sync: Annotated[
        bool, typer.Option(help="Skip all syncing, useful for smoketest")
    ] = False,
):
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y/&m/%d %H:%M:%S",
    )

    if sentry_dsn:
        sentry_sdk.init(dsn=sentry_dsn, environment=get_env(default="None"))

    main_fn = partial(
        main,
        base_deck=deck_name,
        input_dir=input_dir,
        max_deletions_per_run=max_deletions_per_run,
        dry_run=dry_run,
        push_all=push_all,
    )

    if not skip_sync:
        main_fn()
        if watch:
            sleep_seconds = 60
            log.info(f"Sync complete, sleeping for {sleep_seconds} seconds")
            main_fn()


if __name__ == "__main__":
    app()
